from Datasets import *
from copy import copy


class DatasetIterator:
    dataset: Dataset
    batch_size: int
    n_batches: int

    def __init__(self, dataset, batch_size):
        self.dataset = dataset
        self.batch_size = batch_size
        self.n_batches = 0

    def __copy__(self):
        cls = self.__class__
        new_dataset = copy(self.dataset)
        new_iterator = cls(new_dataset, batch_size=self.batch_size)
        return new_iterator

    def fill(self, X, y):
        self.dataset.fill(X, y)
        self.n_batches = int(np.ceil(len(self.dataset) / self.batch_size))
        return self

    def batch(self, iteration):
        if iteration == self.n_batches:
            batch_slice = slice(iteration * self.batch_size, len(self.dataset))
        else:
            batch_slice = slice(iteration * self.batch_size, (iteration + 1) * self.batch_size)
        X_batch, y_batch = self.dataset[batch_slice]
        return X_batch, y_batch

    def __call__(self):
        for it in range(self.n_batches):
            X_batch, y_batch = self.batch(it)
            yield X_batch, y_batch


class SequenceIterator(DatasetIterator):

    def fill(self, X, y):
        self.dataset.fill(X, y)
        self.n_batches = len(self.dataset) - self.batch_size + 1
        return self

    def batch(self, iteration):
        batch_slice = slice(iteration, iteration + self.batch_size)
        X_batch, y_batch = self.dataset[batch_slice]
        return X_batch, y_batch


from grand_lantern.matrix.Matrix import *
from grand_lantern.layers.Activation import ActivationFunction


class Layer:
    parameters: list

    def __init__(self):
        self.parameters = []

    def get_parameters(self):
        return self.parameters

    def forward(self, X):
        pass

    def make_constant(self):
        for param in self.parameters:
            param.require_grad = False
        self.parameters = []
        return self

    def __str__(self):
        return f"Base Layer."


class LinearLayer(Layer):
    W: Matrix
    bias: Matrix
    n_neurons: int
    biased: bool
    activation: ActivationFunction

    def __init__(self, n_neurons, activation, biased=False):
        super().__init__()
        self.n_neurons = n_neurons
        self.activation = activation
        self.biased = biased
        self.W = None
        return

    def initialize_weights(self, n_inputs):
        self.W = Matrix.normal(shape=(n_inputs, self.n_neurons),
                               require_grad=True)
        self.parameters = [self.W]
        if self.biased:
            self.bias = Matrix.normal(shape=(1, self.n_neurons),
                                      require_grad=True)
            self.parameters = [self.W, self.bias]
        return

    def forward(self, X):
        if self.W is None:
            self.initialize_weights(X.shape[1])

        if self.biased:
            return self.activation(X @ self.W + self.bias)

        return self.activation(X @ self.W)

    def __str__(self):
        return f"Layer with n_neurons {self.n_neurons}, " \
               f"biased {self.biased}, " \
               f"activation {self.activation}."


class BatchNormLayer(Layer):
    gamma: Matrix
    beta: Matrix

    def __init__(self):
        super().__init__()
        self.gamma = None
        self.beta = None
        return

    def initialize_weights(self):
        self.gamma = Matrix(1., require_grad=True)
        self.beta = Matrix(0., require_grad=True)
        self.parameters = [self.gamma, self.beta]
        return

    def forward(self, X):
        if (self.gamma is None) or (self.beta is None):
            self.initialize_weights()

        mean = Matrix.mean(X, axis=0, keepdims=True)
        std = Matrix.std(X, axis=0, keepdims=True)

        X_normed = (X - mean) / (std ** 2 + 10e-3)

        return X_normed * self.gamma + self.beta

    def __str__(self):
        return f"Batch Norm Layer."


class Conv2DLayer(LinearLayer):
    kernel_size: np.array([int, int])
    dilation: np.array([int, int])

    def __init__(self, kernel_size, n_channels, activation, dilation=(1, 1), biased=False):
        super().__init__(n_channels, activation, biased)
        self.kernel_size = kernel_size
        self.dilation = dilation
        self.W = None
        return

    def initialize_weights(self, n_inputs):
        self.W = Matrix.normal(shape=(n_inputs, self.n_neurons, self.kernel_size[0], self.kernel_size[1]),
                               require_grad=True)
        self.parameters = [self.W]
        if self.biased:
            self.bias = Matrix.normal(shape=(self.n_neurons),
                                      require_grad=True)
            self.parameters = [self.W, self.bias]
        return

    def forward(self, X):
        if self.W is None:
            self.initialize_weights(X.shape[1])

        if self.biased:
            """
            need to fix
            """
            WX = Matrix.conv2d(X, self.W, self.dilation)
            for c in range(self.n_neurons):
                WX_bias = WX[:, c, :, :] + self.bias[c]

            return self.activation(WX_bias)

        return self.activation(Matrix.conv2d(X, self.W, self.dilation))

    def __str__(self):
        return f"Convolutional layer with kernel {self.kernel_size}, " \
               f"channels {self.n_neurons}, " \
               f"dilation {self.dilation}, " \
               f"biased {self.biased}, " \
               f"activation {self.activation}."


class RNNLayer(Layer):
    Wx: Matrix
    Wh: Matrix
    Wy: Matrix

    bias_h: Matrix
    bias_y: Matrix

    n_inner_neurons: int
    n_out_neurons: int
    biased: bool
    activation: ActivationFunction

    def __init__(self, n_inner_neurons, n_out_neurons, activation, biased=False):
        super().__init__()
        self.n_inner_neurons = n_inner_neurons
        self.n_out_neurons = n_out_neurons
        self.activation = activation
        self.biased = biased
        self.Wx = None
        self.Wh = None
        self.Wy = None
        return

    def initialize_weights(self, n_inputs):
        self.Wx = Matrix.normal(shape=(n_inputs, self.n_inner_neurons), require_grad=True)
        self.Wh = Matrix.normal(shape=(self.n_inner_neurons, self.n_inner_neurons), require_grad=True)
        self.Wy = Matrix.normal(shape=(self.n_inner_neurons, self.n_out_neurons), require_grad=True)
        self.parameters = [self.Wx, self.Wh, self.Wy]

        if self.biased:
            self.bias_h = Matrix.normal(shape=(1, self.n_inner_neurons), require_grad=True)
            self.bias_y = Matrix.normal(shape=(1, self.n_out_neurons), require_grad=True)
            self.parameters = [self.Wx, self.Wh, self.Wy, self.bias_h, self.bias_y]

    def forward(self, X):
        if self.Wx is None:
            self.initialize_weights(X.shape[1])

        h0 = Matrix.zeros(shape=(1, self.n_inner_neurons))
        H = [h0]
        Y = Matrix.zeros(shape=(X.shape[0], self.n_out_neurons))
        for i in range(X.shape[0]):
            if self.biased:
                Xi = X[i].reshape(shape=(1, -1))
                H.append(self.activation(Xi @ self.Wx + H[i] @ self.Wh + self.bias_h))
                Y[i] = H[i + 1] @ self.Wy + self.bias_y
            else:
                Xi = X[i].reshape(shape=(1, -1))
                H.append(self.activation(Xi @ self.Wx + H[i] @ self.Wh))
                Y[i] = H[i + 1] @ self.Wy
        return Y

    def __str__(self):
        return f"RNN Layer with n_inner_neurons {self.n_inner_neurons}, " \
               f"n_out_neurons {self.n_out_neurons}, " \
               f"biased {self.biased}, " \
               f"activation {self.activation}."


class FlattenLayer(Layer):
    input_shape: tuple

    def __init__(self):
        super().__init__()
        self.biased = False
        return

    def forward(self, X):
        self.input_shape = np.array(X.shape)
        X_reshaped = X.reshape(shape=(X.shape[0], np.prod(self.input_shape[1:])))
        return X_reshaped

    def __str__(self):
        return f"Flatten layer."



from grand_lantern.matrix.Matrix import *


class ActivationFunction:

    def __call__(self, X):
        pass

    def __str__(self):
        return f"Base"


class Linear(ActivationFunction):

    def __call__(self, X):
        return X

    def __str__(self):
        return f"Linear"


class Sigmoid(ActivationFunction):

    def __call__(self, X):
        sigmoid = Matrix.sigmoid(X)
        return sigmoid

    def __str__(self):
        return f"Sigmoid"


class Tanh(ActivationFunction):

    def __call__(self, X):
        tanh = Matrix.tanh(X)
        return tanh

    def __str__(self):
        return f"Tanh"


class ReLU(ActivationFunction):

    def __call__(self, X):
        return X * (Matrix.sign(X) + 1) / 2

    def __str__(self):
        return f"ReLU"


class LReLU(ActivationFunction):
    alpha: float

    def __init__(self, alpha=0.01):
        self.alpha = alpha

    def __call__(self, X):
        return X * ((1 + Matrix.sign(X)) / 2 + self.alpha * (1 + Matrix.sign(-X)) / 2)

    def __str__(self):
        return f"LReLU"


class SoftMax(ActivationFunction):

    def __call__(self, X):
        return Matrix.safe_softmax(X)

    def __str__(self):
        return f"SoftMax"



from grand_lantern.matrix.Matrix import *


class Metric:

    def __call__(self, y_true, y_pred):
        pass

    def __str__(self):
        return f"Base"


class Accuracy(Metric):

    def __call__(self, y_true, y_pred):
        yt = np.argmax(y_true.value, axis=1)
        yp = np.argmax(y_pred.value, axis=1)
        return len(np.where(yt == yp)[0]) / len(yt)

    def __str__(self):
        return f"Accuracy"


class MSEMetric(Metric):

    def __call__(self, y_true, y_pred):
        summ = np.sum((y_true.value - y_pred.value) ** 2)
        return 1 / len(y_true.value) * summ

    def __str__(self):
        return f"MSE"


class R2(Metric):

    def __call__(self, y_true, y_pred):
        ss_res = np.sum((y_true.value - y_pred.value) ** 2)
        ss_tot = np.sum((y_true.value - np.mean(y_true.value)) ** 2)
        return 1 - ss_res / ss_tot

    def __str__(self):
        return f"R2"



from grand_lantern.matrix.Matrix import *


class Optimizer:

    def optimize(self, parameters, gradients):
        pass


class SGD(Optimizer):
    learning_rate: float

    def __init__(self, learning_rate):
        self.learning_rate = learning_rate
        return

    def optimize(self, parameters, gradients):
        for param in parameters:
            param.value -= self.learning_rate * gradients[param]
        return self


class NAD(SGD):
    inertia_moment: float
    last_gradients: dict

    def __init__(self, learning_rate, inertia_moment):
        super().__init__(learning_rate)
        self.inertia_moment = inertia_moment
        self.last_gradients = {}
        return

    def optimize(self, parameters, gradients):
        for param in parameters:
            if param not in self.last_gradients:
                self.last_gradients[param] = self.learning_rate * gradients[param]
            else:
                self.last_gradients[param] = self.learning_rate * gradients[param] + self.inertia_moment * \
                                             self.last_gradients[param]
            param.value -= self.learning_rate * self.last_gradients[param]
        return self


class Adagrad(SGD):
    sum_square_update: np.array

    def __init__(self, learning_rate):
        super().__init__(learning_rate)
        self.sum_square_update = {}
        return

    def optimize(self, parameters, gradients):
        for param in parameters:
            if param not in self.sum_square_update:
                self.sum_square_update[param] = gradients[param] ** 2
            else:
                self.sum_square_update[param] += gradients[param] ** 2

            param.value -= self.learning_rate / (np.sqrt(self.sum_square_update[param] + 10e-10)) * gradients[param]
        return self


class Adam(NAD, Adagrad):
    beta1: float
    beta2: float
    iteration: int

    def __init__(self, learning_rate, beta1=0.9, beta2=0.99):
        self.learning_rate = learning_rate
        self.beta1 = beta1
        self.beta2 = beta2
        self.sum_square_update = {}
        self.last_gradients = {}
        self.iteration = 0
        return

    def optimize(self, parameters, gradients):
        self.iteration += 1

        for param in parameters:
            if param not in self.last_gradients:
                self.last_gradients[param] = (1 - self.beta1) * gradients[param]
                self.sum_square_update[param] = (1 - self.beta2) * gradients[param] ** 2
            else:
                self.last_gradients[param] = (1 - self.beta1) * gradients[param] + self.beta1 * self.last_gradients[
                    param]
                self.sum_square_update[param] = (1 - self.beta2) * gradients[param] ** 2 + self.beta2 * \
                                                self.sum_square_update[param]

            new_gradient = self.last_gradients[param] / (1 - self.beta1 ** self.iteration)
            new_sum_square_update = self.sum_square_update[param] / (1 - self.beta2 ** self.iteration)

            param.value -= self.learning_rate / (np.sqrt(new_sum_square_update + 10e-10)) * new_gradient
        return self



from grand_lantern.matrix.Matrix import *


class Dataset():

    def __init__(self):
        return

    def fill(self, X, y):
        pass

    def __len__(self):
        pass

    def __getitem__(self, idx):
        pass


class TableDataset(Dataset):
    X: Matrix
    y: Matrix

    def __init__(self):
        super().__init__()
        return

    def fill(self, X, y):
        if X.shape[0] != y.shape[0]:
            raise ValueError
        self.X = Matrix(X)
        self.y = Matrix(y)
        return self

    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, idx):
        return self.X[idx], self.y[idx]


class ImageDataset(Dataset):
    X: Matrix
    y: Matrix

    def __init__(self):
        super().__init__()
        return

    def fill(self, X, y):
        if X.shape[0] != y.shape[0]:
            raise ValueError
        if X.ndim == 3:
            X = X.reshape(X.shape[0], X.shape[1], X.shape[2], 1)
        self.X = Matrix(X)
        self.y = Matrix(y)
        return self

    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, idx):
        return self.X[idx], self.y[idx]


from setuptools import setup, find_packages

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "ReadMe.md").read_text()

setup(
  name='grandlantern',
  version='0.0.7',
  author='grandyarl',
  author_email='grand_yarl@mail.com',
  url='https://github.com/grand-yarl/grand_lantern',
  description='This is the small library for deep learning.',
  long_description=long_description,
  long_description_content_type='text/markdown',
  packages=find_packages(),
  classifiers=[
    'Programming Language :: Python :: 3.9',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent'
  ],
  keywords='deeplearn',
  project_urls={
    'GitHub': 'https://github.com/grand-yarl/grand_lantern'
  },
  python_requires='>=3.6'
)


from grand_lantern.matrix.Matrix import *


class Loss:

    def __call__(self, y_true, y_pred):
        pass

    def __str__(self):
        return f"Base"


class MSELoss(Loss):

    def __call__(self, y_true, y_pred):
        return (y_pred - y_true) ** 2

    def __str__(self):
        return f"MSE"


class CrossEntropy(Loss):

    def __call__(self, y_true, y_pred):
        summ = -1 * Matrix.sum(y_true * Matrix.log(y_pred + 10e-5), axis=1)
        return summ

    def __str__(self):
        return f"CrossEntropy"



import numpy as np
from copy import copy
from collections import defaultdict


class Matrix:

    value: np.array
    local_gradients: list
    require_grad: bool
    shape: tuple

    def __init__(self, array, local_gradients=None, require_grad=False):
        if local_gradients is None:
            local_gradients = []
        if isinstance(array, Matrix):
            self.value = array.value
            self.local_gradients = array.local_gradients
        else:
            self.value = np.array(array)
            self.local_gradients = local_gradients
        self.shape = self.value.shape
        self.require_grad = require_grad

    def __add__(self, other):

        def compute_gradient(grad, target):
            while np.prod(grad.shape) > np.prod(target.shape):
                grad = np.sum(grad, axis=0)
            return grad

        if not isinstance(other, Matrix):
            other = Matrix(other)

        new_value = self.value + other.value
        new_local_gradients = []
        new_require_grad = self.require_grad or other.require_grad

        if self.require_grad:
            new_local_gradients.append((self, lambda x: compute_gradient(x, self), 'add'))
        if other.require_grad:
            new_local_gradients.append((other, lambda x: compute_gradient(x, other), 'add'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    def __sub__(self, other):
        def compute_gradient(grad, target):
            while np.prod(grad.shape) > np.prod(target.shape):
                grad = np.sum(grad, axis=0)
            return grad

        if not isinstance(other, Matrix):
            other = Matrix(other)

        new_value = self.value - other.value
        new_local_gradients = []
        new_require_grad = self.require_grad or other.require_grad

        if self.require_grad:
            new_local_gradients.append((self, lambda x: compute_gradient(x, self), 'sub'))
        if other.require_grad:
            new_local_gradients.append((other, lambda x: -compute_gradient(x, other), 'sub'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    def __mul__(self, other):
        def compute_gradient(grad, target):
            while np.prod(grad.shape) > np.prod(target.shape):
                grad = np.sum(grad, axis=0)
            return grad

        if not isinstance(other, Matrix):
            other = Matrix(other)

        new_value = self.value * other.value
        new_local_gradients = []
        new_require_grad = self.require_grad or other.require_grad

        if self.require_grad:
            new_local_gradients.append((self, lambda x: compute_gradient(x * other.value, self), 'mul'))
        if other.require_grad:
            new_local_gradients.append((other, lambda x: compute_gradient(x * self.value, other), 'mul'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    def __truediv__(self, other):
        def compute_gradient(grad, target):
            while np.prod(grad.shape) > np.prod(target.shape):
                grad = np.sum(grad, axis=0)
            return grad

        if not isinstance(other, Matrix):
            other = Matrix(other)

        new_value = self.value / other.value
        new_local_gradients = []
        new_require_grad = self.require_grad or other.require_grad

        if self.require_grad:
            new_local_gradients.append((self, lambda x: compute_gradient(x / other.value, self), 'div'))
        if other.require_grad:
            new_local_gradients.append(
                (other, lambda x: -compute_gradient(x * self.value / (other.value ** 2), other), 'div'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    def __radd__(self, other):
        if not isinstance(other, Matrix):
            other = Matrix(other)
        return other + self

    def __rsub__(self, other):
        if not isinstance(other, Matrix):
            other = Matrix(other)
        return other - self

    def __rmul__(self, other):
        if not isinstance(other, Matrix):
            other = Matrix(other)
        return other * self

    def __rtruediv__(self, other):
        if not isinstance(other, Matrix):
            other = Matrix(other)
        return other / self

    def __neg__(self):
        new_value = -self.value
        new_local_gradients = []
        new_require_grad = self.require_grad

        if self.require_grad:
            new_local_gradients.append((self, lambda x: -x, 'neg'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    def __matmul__(self, other):
        if not isinstance(other, Matrix):
            other = Matrix(other)

        new_value = self.value @ other.value
        new_local_gradients = []
        new_require_grad = self.require_grad or other.require_grad

        if self.require_grad:
            new_local_gradients.append((self, lambda x: x @ other.value.T, 'matmul'))
        if other.require_grad:
            new_local_gradients.append((other, lambda x: self.value.T @ x, 'matmul'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    def __pow__(self, power):
        new_value = self.value ** power
        new_local_gradients = []
        new_require_grad = self.require_grad

        if self.require_grad:
            new_local_gradients.append((self, lambda x: x * power * (self.value ** (power - 1)), 'pow'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    def reshape(self, shape):
        old_shape = self.shape
        new_value = self.value.reshape(shape)
        new_local_gradients = []
        new_require_grad = self.require_grad

        if self.require_grad:
            new_local_gradients.append((self, lambda x: x.reshape(old_shape), 'reshape'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def zeros(cls, shape, require_grad=False):
        new_value = np.zeros(shape)
        return Matrix(new_value, require_grad=require_grad)

    @classmethod
    def normal(cls, shape, mean=0, std=1, require_grad=False):
        new_value = np.random.normal(loc=mean, scale=std, size=shape)
        return Matrix(new_value, require_grad=require_grad)

    @classmethod
    def uniform(cls, shape, low=-1, high=1, require_grad=False):
        new_value = np.random.uniform(low=low, high=high, size=shape)
        return Matrix(new_value, require_grad=require_grad)

    @classmethod
    def sin(cls, obj):
        new_value = np.sin(obj.value)
        new_local_gradients = []
        new_require_grad = obj.require_grad

        if obj.require_grad:
            new_local_gradients.append((obj, lambda x: x * np.cos(obj.value), 'sin'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def cos(cls, obj):
        new_value = np.cos(obj.value)
        new_local_gradients = []
        new_require_grad = obj.require_grad

        if obj.require_grad:
            new_local_gradients.append((obj, lambda x: -x * np.sin(obj.value), 'cos'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def exp(cls, obj):
        new_value = np.exp(obj.value)
        new_local_gradients = []
        new_require_grad = obj.require_grad

        if obj.require_grad:
            new_local_gradients.append((obj, lambda x: x * new_value, 'exp'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def log(cls, obj):
        new_value = np.log(obj.value)
        new_local_gradients = []
        new_require_grad = obj.require_grad

        if obj.require_grad:
            new_local_gradients.append((obj, lambda x: x / obj.value, 'log'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def sqrt(cls, obj):
        new_value = np.sqrt(obj.value)
        new_local_gradients = []
        new_require_grad = obj.require_grad

        if obj.require_grad:
            new_local_gradients.append((obj, lambda x: x / (2 * obj.value + 10e-5), 'sqrt'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def sum(cls, obj, axis=None, keepdims=False):
        if not keepdims:  # Не хотим сохранить размерность
            if axis is not None:
                new_value = np.sum(obj.value, axis=axis)
                new_local_gradients = []
                new_require_grad = obj.require_grad

                if obj.require_grad:
                    new_local_gradients.append(
                        (obj, lambda x: np.expand_dims(np.array(x), axis=axis) * np.ones_like(obj.value), 'sum')
                    )
                return Matrix(new_value, new_local_gradients, new_require_grad)

            else:
                new_value = np.sum(obj.value, axis=axis)
                new_local_gradients = []
                new_require_grad = obj.require_grad

                if obj.require_grad:
                    new_local_gradients.append(
                        (obj, lambda x: x * np.ones_like(obj.value), 'sum')
                    )
                return Matrix(new_value, new_local_gradients, new_require_grad)

        else:  # Хотим сохранить размерность
            new_value = np.sum(obj.value, axis=axis, keepdims=True) * np.ones_like(obj.value)
            new_local_gradients = []
            new_require_grad = obj.require_grad

            if obj.require_grad:
                new_local_gradients.append(
                    (obj, lambda x: x, 'sum')
                )
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def mean(cls, obj, axis=None, keepdims=False):
        if not keepdims:
            if axis is not None:
                new_value = np.sum(obj.value, axis=axis)
                new_local_gradients = []
                new_require_grad = obj.require_grad

                if obj.require_grad:
                    n_elements = np.prod(np.array(obj.shape[axis]))
                    new_local_gradients.append(
                        (obj, lambda x: (np.expand_dims(np.array(x), axis=axis) + np.zeros(obj.shape)) / n_elements,
                         'mean')
                    )
                return Matrix(new_value, new_local_gradients, new_require_grad)

            else:
                new_value = np.mean(obj.value, axis=axis)
                new_local_gradients = []
                new_require_grad = obj.require_grad

                if obj.require_grad:
                    n_elements = np.prod(np.array(obj.shape))
                    new_local_gradients.append(
                        (obj, lambda x: (x + np.zeros(obj.shape)) / n_elements, 'mean')
                    )
                return Matrix(new_value, new_local_gradients, new_require_grad)

        else:
            new_value = np.mean(obj.value, axis=axis, keepdims=True) * np.ones_like(obj.value)
            new_local_gradients = []
            new_require_grad = obj.require_grad

            if obj.require_grad:
                n_elements = None
                if axis is not None:
                    n_elements = np.prod(np.array(obj.shape[axis]))
                else:
                    n_elements = np.prod(np.array(obj.shape))
                new_local_gradients.append(
                    (obj, lambda x: x / n_elements, 'mean')
                )
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def std(cls, obj, axis=None, keepdims=False):
        mean = Matrix.mean(obj, axis=axis, keepdims=keepdims)
        sub = obj - mean
        square_sub = sub ** 2
        sum_square = Matrix.mean(square_sub, axis=axis, keepdims=keepdims)
        std = Matrix.sqrt(sum_square)
        return std

    @classmethod
    def sigmoid(cls, obj):
        new_value = 1 / (1 + np.exp(-1 * obj.value))
        new_local_gradients = []
        new_require_grad = obj.require_grad

        if obj.require_grad:
            new_local_gradients.append((obj, lambda x: x * new_value * (1 - new_value), 'sigmoid'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def sign(cls, obj):
        new_value = np.sign(obj.value)
        return Matrix(new_value)

    @classmethod
    def tanh(cls, obj):
        new_value = np.tanh(obj.value)
        new_local_gradients = []
        new_require_grad = obj.require_grad

        if obj.require_grad:
            new_local_gradients.append((obj, lambda x: x * (1 + new_value) * (1 - new_value), 'tanh'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def softmax(cls, obj, axis=-1):

        def compute_gradient(grad):
            prime = np.zeros(obj.shape)
            for i in range(obj.shape[0]):
                s = np.array([new_value[i]])
                g = np.array([grad[i]])
                prime[i] = -s * np.sum(s * g, axis=1) + s * g
            return prime

        new_value = np.exp(obj.value) / np.sum(np.exp(obj.value), axis=axis, keepdims=True)
        new_local_gradients = []
        new_require_grad = obj.require_grad

        if obj.require_grad:
            new_local_gradients.append((obj, lambda x: compute_gradient(x), 'softmax'))
        return Matrix(new_value, new_local_gradients, new_require_grad)

    @classmethod
    def safe_softmax(cls, obj, axis=-1):
        sub = Matrix(obj.value.max(axis=1, keepdims=True))
        return Matrix.softmax(obj - sub, axis=axis)

    @classmethod
    def conv2d(cls, matrix, kernel, dilation):

        def convolution(matrix, kernel, dilation):
            x_steps = (matrix.shape[2] - kernel.shape[2] + 1) // dilation[0]
            y_steps = (matrix.shape[3] - kernel.shape[3] + 1) // dilation[1]
            conv = np.zeros((matrix.shape[0], kernel.shape[1], x_steps, y_steps))
            for i in range(x_steps):
                for j in range(y_steps):
                    for c in range(kernel.shape[1]):
                        x_slice = slice(i * dilation[0], i * dilation[0] + kernel.shape[2])
                        y_slice = slice(j * dilation[1], j * dilation[1] + kernel.shape[3])
                        conv[:, c, i, j] = np.sum(matrix[:, :, x_slice, y_slice] * kernel[:, c, :, :])
            return conv

        def grad_dilate(grad, dilation):
            new_grad = np.zeros(
                (grad.shape[0], grad.shape[1], dilation[0] * grad.shape[2], dilation[1] * grad.shape[3]))
            new_grad[:, :, ::dilation[0], ::dilation[1]] = grad
            return new_grad

        def grad_pad(grad, matrix):
            pad_x = matrix.shape[2] - grad.shape[2]
            pad_y = matrix.shape[3] - grad.shape[3]
            return np.pad(grad, ((0, 0), (0, 0), (pad_x, pad_x), (pad_y, pad_y)))

        new_value = convolution(matrix.value, kernel.value, dilation)
        new_local_gradients = []
        new_require_grad = matrix.require_grad or kernel.require_grad

        if matrix.require_grad:
            new_local_gradients.append(
                (matrix, lambda x: convolution(
                    grad_pad(grad_dilate(x, dilation), matrix),
                    kernel.value[:, :, ::-1, ::-1].reshape(kernel.shape[1], kernel.shape[0], kernel.shape[2],
                                                           kernel.shape[3]),
                    dilation=(1, 1))[:, :, ::-1, ::-1],
                 'conv2d')
            )
        if kernel.require_grad:
            new_local_gradients.append(
                (kernel, lambda x: convolution(
                    matrix.value.reshape(matrix.shape[1], matrix.shape[0], matrix.shape[2], matrix.shape[3]),
                    grad_dilate(x[:, :, ::-1, ::-1], dilation),
                    dilation=(1, 1)),
                 'conv2d')
            )
        return Matrix(new_value, new_local_gradients, new_require_grad)

    def backward(self):

        gradients = defaultdict(lambda: 0)

        def compute_gradients(matrix, before_grads):
            if matrix.local_gradients:
                for (child, child_gradients_func, operation) in matrix.local_gradients:
                    new_child_grad = child_gradients_func(before_grads)
                    compute_gradients(child, new_child_grad)
                    gradients[child] += new_child_grad

        compute_gradients(self, np.ones(self.shape))

        return gradients

    def __getitem__(self, idx):
        return Matrix(self.value[idx])

    def __setitem__(self, key, item):
        if isinstance(item, Matrix):
            self.value[key] = item.value
            if item.require_grad:
                self.require_grad = True
                self.local_gradients.append((item, lambda x: x[key].reshape(1, -1), 'setitem'))
        else:
            self.value = np.array(item)
        return

    def __repr__(self):
        return f"{self.value}"

    def __str__(self):
        return f"{self.value}"



from grand_lantern.layers.Layers import *
from grand_lantern.metrics import *
from grand_lantern.optimizers import *
from grand_lantern.dataiterators.DatasetIterators import *


class model():
    layers: list[Layer]
    n_epochs: int
    dataset_it: DatasetIterator
    loss_fn: Loss
    metric_fn: Metric
    opimizer: Optimizer
    fit_error: np.array
    val_error: np.array
    parameters = []

    def __init__(self, n_epochs, dataset_iterator, loss_function, metric_function, optimizer):
        self.layers = []
        self.n_epochs = n_epochs
        self.dataset_it = dataset_iterator
        self.loss_fn = loss_function
        self.metric_fn = metric_function
        self.optimizer = optimizer
        self.parameters = []
        return

    def add_layer(self, layer):
        self.layers.append(layer)
        return

    def forward(self, input_model):
        self.parameters = []
        current_input = input_model
        for layer in self.layers:
            current_input = layer.forward(current_input)
            self.parameters += layer.get_parameters()
        return current_input

    def zero_grad(self):
        for parameter in self.parameters:
            parameter.local_gradients = []
        return

    def fit(self, X, y, X_val=None, y_val=None):

        train_dataset_iterator = copy(self.dataset_it)
        train_dataset_iterator.fill(X, y)
        n_batches_train = train_dataset_iterator.n_batches
        self.fit_error = np.zeros((self.n_epochs))

        val_dataset_iterator = None
        n_batches_val = 0
        if (X_val is not None) and (y_val is not None):
            val_dataset_iterator = copy(self.dataset_it)
            val_dataset_iterator.fill(X_val, y_val)
            n_batches_val = val_dataset_iterator.n_batches
            self.val_error = np.zeros((self.n_epochs))

        for epoch in range(self.n_epochs):

            sum_loss_train = 0
            sum_metric_train = 0

            for (X_batch, y_batch) in train_dataset_iterator():
                self.zero_grad()
                y_pred = self.forward(X_batch)

                loss = self.loss_fn(y_batch, y_pred)
                metric = self.metric_fn(y_batch, y_pred)
                gradients = loss.backward()
                self.optimizer.optimize(self.parameters, gradients)
                # print(Matrix.mean(loss))

                sum_loss_train += np.mean(loss.value)
                sum_metric_train += metric

            loss_train = sum_loss_train / n_batches_train
            metric_train = sum_metric_train / n_batches_train

            loss_msg = f"Epoch {epoch + 1:>4d}: Train {self.loss_fn}: {loss_train:==7f} "
            metric_msg = f"Epoch {epoch + 1:>4d}: Train {self.metric_fn}: {metric_train:==7f} "
            self.fit_error[epoch] = loss_train

            if (X_val is not None) and (y_val is not None):

                sum_loss_val = 0
                sum_metric_val = 0

                for (X_batch, y_batch) in val_dataset_iterator():
                    self.zero_grad()
                    y_pred = self.forward(X_batch)

                    loss = self.loss_fn(y_batch, y_pred)
                    metric = self.metric_fn(y_batch, y_pred)

                    sum_loss_val += np.mean(loss.value)
                    sum_metric_val += metric

                loss_val = sum_loss_val / n_batches_val
                metric_val = sum_metric_val / n_batches_val

                loss_msg += f" Test {self.loss_fn}: {loss_val:==7f} "
                metric_msg += f" Test {self.metric_fn}: {metric_val:==7f} "
                self.val_error[epoch] = loss_val
            print(loss_msg)
            print(metric_msg)
            print(len(metric_msg) * "-")
        return self

    def predict(self, X):
        X = Matrix(X)
        y_pred = self.forward(X).value
        return y_pred

    def make_constant_layers(self):
        for layer in self.layers:
            layer.make_constant()
        return self

    def __str__(self):
        model_str = ""
        for layer in self.layers:
            model_str += (str(layer) + "\n")
        return model_str
